;;;; Copyright 2005 Juri Pakaste
;;;; 
;;;; This file is part of Lukutoukka.
;;;; 
;;;; Lukutoukka is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;; 
;;;; Lukutoukka is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with Lukutoukka; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(define-module (lukutoukkalib file)
  #:use-module (lukutoukkalib srfi srfi-56)
  #:use-module (lukutoukkalib book)
  #:use-module (lukutoukkalib utils)
  #:use-module (lukutoukkalib parsers common)
  #:use-module (oop goops)
  #:export (<book-file-builder>
            create-new-file
            create-file-reader
            initialize-file
            save-word
            end-words
            read-next-word
            read-word-at-point
            parse-file-to-file-backed-book))

(define +file-format-version+ 0)
(define +file-format-marker+ "lukutoukka")
(define +toc-pointer-location+ 12)

; book-file-builder: write out the files
(define-class <book-file-builder> ()
  (word-offsets #:init-form '() #:accessor word-offsets)
  (file-name #:init-keyword #:file-name #:getter file-name)
  (file-port #:init-keyword #:file-port #:getter file-port)
  (initialized #:init-form #f #:accessor initialized)
  (finally-close-port #:init-form #t #:accessor finally-close-port))

(define-method (create-new-file (name <string>))
  (let ((builder (make <book-file-builder> #:file-name name 
                       #:file-port (open-binary-output-file name))))
    (initialize-file builder)))

(define-method (create-new-file (name <string>) (port <output-port>))
  (let ((builder (make <book-file-builder> #:file-name name
                       #:file-port port)))
    (set! (finally-close-port builder) #f)
    (initialize-file builder)))

(define-method (initialize-file (builder <book-file-builder>))
  (let ((port (file-port builder)))
    (display +file-format-marker+ port)
                                        ; file format version, two bytes
    (write-network-uint16 +file-format-version+ port) 
    (write-network-uint64 0 port)) ; location of the toc, will overwrite later
  (set! (initialized builder) #t)
  builder)

(define-method (save-word (builder <book-file-builder>) word)
  (assert (initialized builder) "The file has not been initialized!")
  (assert (or (string? word) 
              (and (pair? word) 
                   (string? (car word)) 
                   (number? (cadr word))))
          "Argument 2 must be either a word or a (word number) pair")
  (let ((current-offsets (word-offsets builder)))
    (set! (word-offsets builder) (cons (ftell (file-port builder)) 
                                       current-offsets))
    (write word (file-port builder))))

(define-method (end-words (builder <book-file-builder>))
  (assert (initialized builder) "The file has not been initialized!")
  (let* ((port (file-port builder))
         (toc-location (ftell port)))
    (display (list->vector (reverse! (word-offsets builder))) port)
                                        ;seek to the point where the 
                                        ;location of the toc should be written
    (seek port +toc-pointer-location+ SEEK_SET)
    (write-network-uint64 toc-location port)
    (when (finally-close-port builder)
          (close-port port)))
  builder)

; book-file-reader: read files
(define-class <book-file-reader> ()
  (file-name #:init-keyword #:file-name #:getter file-name)
  (file-port #:init-keyword #:file-port #:getter file-port)
  (toc #:init-form #f #:accessor file-toc)
  (point #:init-form 0 #:accessor toc-point))

(define-method (create-file-reader (name <string>))
  (let ((r (make <book-file-reader> 
             #:file-name name 
             #:file-port (open-binary-input-file name))))
    (check-file-and-read-toc r)
    r))

(define-method (create-file-reader (port <input-port>))
  (let ((r (make <book-file-reader>
             #:file-port port)))
    (check-file-and-read-toc r)
    r))

(define (read-characters port n)
  (let loop ((characters '())
             (n n))
    (if (and (> n 0) (char-ready? port))
        (loop (cons (read-char port) characters) (1- n))
        (reverse! characters))))

(define-method (check-file-and-read-toc (reader <book-file-reader>))
  (let ((port (file-port reader)))
    (assert (equal? +file-format-marker+ 
                    (apply string (read-characters port 10)))
            "File doesn't appear to be an lukutoukka file")
    (assert (equal? +file-format-version+ 
                    (read-network-uint16 port))
            "Wrong file format version")
    (seek port (read-network-uint64 port) SEEK_SET)
    (set! (file-toc reader) (read port))))

(define-method (read-next-word (reader <book-file-reader>))
  (if (= (toc-point reader) (vector-length (file-toc reader)))
      #f
      (let ((port (file-port reader)))
        (seek port (vector-ref (file-toc reader) (toc-point reader)) SEEK_SET)
        (set! (toc-point reader) (1+ (toc-point reader)))
        (read port))))

(define-method (read-word-at-point (reader <book-file-reader>) 
                                   (point <integer>))
  (if (not (<= 0 
               point 
               (1- (vector-length (file-toc reader)))))
      #f
      (let ((port (file-port reader)))
        (seek port (vector-ref (file-toc reader) point) SEEK_SET)
        (read port))))

(define-class <file-backed-book> (<book>)
  (file-reader #:init-keyword #:file-reader #:getter file-reader))

(define-method (get-next-word (book <file-backed-book>))
  (let ((word-index (min (- (number-of-words book) 1)
                         (max 0 ((if (eq? (direction book) 'forward) + -) 
                                 (position book)
                                 1)))))
    (set! (position book) word-index)
    (read-word-at-point (file-reader book) word-index)))

(define-method (number-of-words (book <file-backed-book>))
  (vector-length (file-toc (file-reader book))))

(define (parse-file-to-file-backed-book filename)
  ; the mkstemp! input string can't be a literal because it gets modified
  (unless (access? filename (logior F_OK R_OK))
          (throw 'no-such-file-or-no-permissions))
  (let* ((temp-file-name (string-append "/tmp/lukutoukka" "-XXXXXX"))
         (temp-file-port (mkstemp! temp-file-name))
         (builder (make <book-file-builder> 
                    #:file-name temp-file-name
                    #:file-port temp-file-port)))
    (initialize-file builder)
    (parse-file-for-consumer filename (lambda (word) (save-word builder word)))
    (end-words builder)
    (make <file-backed-book> 
      #:file-reader (create-file-reader temp-file-name)
      #:book-name filename)))
