;;;; Copyright 2005 Juri Pakaste
;;;; 
;;;; This file is part of Lukutoukka.
;;;; 
;;;; Lukutoukka is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;; 
;;;; Lukutoukka is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with Lukutoukka; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(define-module (lukutoukkalib parsers common)
  #:use-module (lukutoukkalib parsers defs)
  #:use-module (lukutoukkalib parsers plain)
  #:use-module (lukutoukkalib utils)
  #:use-module (oop goops)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:export (add-pauses get-parser-for-file parse-file-for-consumer parse-file-to-vector))

(define (get-parser-for-file file)
  (make <plain-text-source> #:port (open-input-file file)))

(define (parse-file-for-consumer file consumer)
  (let ((parser (get-parser-for-file file)))
    (when parser
          (parse-port parser (lambda (word) (consumer (word-with-pause word)))))))

(define (parse-file-to-vector file)
  (define (get-word-list parser)
    (let ((words-hash (make-hash-table 7919))
          (words-list '()))
      (parse-port parser 
                  (lambda (value)
                    (if value
                        (let ((previously-hashed 
                               (hash-ref words-hash value)))
                          (if previously-hashed
                              (begin
                                (set! words-list 
                                      (cons previously-hashed words-list)))
                              (begin
                                (let ((paused-word (word-with-pause value)))
                                  (hash-set! words-hash value paused-word)
                                  (set! words-list (cons paused-word 
                                                         words-list)))))))))
      (reverse! words-list)))
  (let ((parser (get-parser-for-file file)))
    (when parser
          (list->vector (get-word-list parser)))))

(define *pause-table*
  '(((#\. #\! #\?) . 2)
    ((#\-) . 1.6)
    ((paragraph) . 3)
    ((#\,) . 1.6)
    ((#\( #\)) . 1.3)))

(define (get-pause ob)
  (define pause-elements (lambda (pair)
                           (car pair)))
  (define pause-amount (lambda (pair)
                         (cdr pair)))
  (let find-pause ((pauses *pause-table*))
    (if (not (null-list? pauses))
        (let ((head (car pauses))
              (tail (cdr pauses)))
          (if (member ob (pause-elements head))
              (pause-amount head)
              (find-pause tail)))
        1)))

(define (word-with-pause word)
  (let* ((pause-amount 
          (get-pause (string-ref word (1- (string-length word)))))
         (paused-word 
          (if (= 1 pause-amount)
              word
              (list word pause-amount))))
    paused-word))

(define (add-pauses parsed)
  "Add pauses to a list of words, checking the last
letter of each word and looking up the appropriate pause value
from *pause-table*.  Each word that requires a pause after it
becomes a pair of word and a multiplier for the default pause
time.
@var{parsed} is a list of strings."
  (apply append! 
         (map (lambda (word) 
                (let ((pause-amount 
                       (get-pause 
                        (string-ref word (- (string-length word) 1)))))
                  (list 
                   (if (= 1 pause-amount)
                       word
                       (list word pause-amount)))))
              parsed)))
