;;;; Copyright 2005 Juri Pakaste
;;;; 
;;;; This file is part of Lukutoukka.
;;;; 
;;;; Lukutoukka is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;; 
;;;; Lukutoukka is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with Lukutoukka; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(define-module (lukutoukkalib ui mainwindow)
  #:use-module (lukutoukkalib ui error)
  #:use-module (lukutoukkalib utils)
  #:use-module (lukutoukkalib parsers common)
  #:use-module ((lukutoukkalib book)
                :renamer (symbol-prefix-proc 'book:))
  #:use-module (lukutoukkalib bookmarks)
  #:use-module (lukutoukkalib file)
  #:use-module (gnome gnome)
  #:use-module (gnome gtk)
  #:use-module (gnome glade)
  #:use-module (gnome glib)
  #:use-module (gnome pango)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-1)
  #:export (show-main-window))

(define +window-name+ "reader_window")
(define +menu-container-name+ "menu_container")
(define +word-display-name+ "word_label")
(define +speed-scale-name+ "speed_scale")
(define +control-container-name+ "control_container")
(define +word-count-label-name+ "total_word_count_label")
(define +word-number-spin+ "current_word_number_spin")
(define +about-window-name+ "lukutoukka_about")
(define +preferences-window-name+ "lukutoukka_preferences")
(define +menu-path+ "/ui/MainMenu")

; main window class and a variable to hold the instance

(define-class <main-window> ()
  (widget #:init-argument #:widget #:getter widget)
  (book #:init-value #f #:accessor open-book)
  (word-display #:init-value '() #:accessor word-display)
  (speed-scale #:init-value '() #:accessor speed-scale)
  (control-container #:init-value '() #:accessor control-container)
  (word-count-label #:init-value '() #:accessor word-count-label)
  (word-number-spin #:init-value '() #:accessor word-number-spin)
  (paused #:init-value #t #:accessor paused))

(define *window* #f)
(define *about-window* #f)
(define *preferences-window* #f)

(define (save-bookmark name position)
  (let loop ((times 10)
             (sleep-time (* 1000 (random 10))))
    (if (zero? times)
        (show-error-dialog 
         "Sorry, could not save bookmarks due to a held lock.")
        (catch 'bookmarks-store-failed
               (lambda ()
                 (store-bookmark name position))
               (lambda (key)
                 (usleep sleep-time)
                 (loop (- times 1) (* sleep-time 2)))))))

(define (open-book-file filename window)
  (let ((old-book (open-book *window*)))
    (if old-book
        (store-bookmark (book:book-name old-book) (book:position old-book))))
  (set! filename (absolute-path filename))
  (catch 'no-such-file-or-no-permissions
         (lambda ()
           (let* ((book (parse-file-to-file-backed-book filename)))
                                        ;  (let* ((book-words (parse-file-to-vector filename))
                                        ;         (book (make book:<book> #:words book-words #:book-name filename)))

             (set! (open-book window) book)
             (set-sensitive (speed-scale window) #t)
             (set-sensitive (control-container window) #t)
             (set-range (word-number-spin window) 
                        0 
                        (book:number-of-words book))
             (set-text (word-count-label window) 
                       (format #f "/ ~a" (book:number-of-words book)))
             (let ((bookmark (find-bookmark filename)))
               (when bookmark
                     (set! (book:position book) bookmark)
                     (set-value (word-number-spin window) bookmark)))))
         (lambda (key)
           (show-error-dialog (format #f "No such file or no permission to access:~%~a" filename)))))

(define (open-file-if-argument arguments window)
  (let* ((filtered (filter (lambda (arg) (not (string-prefix? "-" arg))) 
                          (cdr arguments)))
         (number-of-args (length filtered)))
    
    (cond 
     ((= 1 number-of-args)
      (open-book-file (car filtered) window))
     ((< 1 number-of-args)
      (format #t "Invalid number of arguments: ~a~%" number-of-args)))))

(define (quit-program)
  (let ((book (open-book *window*)))
    (if book
        (store-bookmark (book:book-name book) (book:position book))))
  (gtk-main-quit))

;; signal guard for the spin button signal, we change it programmatically and
;; don't want to handle the value-changed signal in that situation
(define *word-number-spin-guard* #f)

;; Event handlers

;; These are at the moment connected with the autoconnect thing, but
;; maybe it isn't the best solution - now they have to use the lexical
;; variable *window* to find the window. I'd really prefer for them to
;; get it as an argument.

; time slider change
(define *restart-timer* 0)

(define (word-timer timecount dir timeout)
  (define (timeout-callback)
    (word-timer timecount dir timeout))
  (unless (> *restart-timer* timecount)
          (if (not (paused *window*))
              (let* ((book (open-book *window*))
                     (next-pair ((lambda (p) (if (pair? p) p (list p 1))) 
                                 (book:get-next-word book)))
                     (next-word (car next-pair))
                     (next-timeout (cadr next-pair)))
                (set! (book:direction book) dir)
                (set-text (word-display *window*) next-word)
                (with-guard *word-number-spin-guard*
                            (set-value (word-number-spin *window*) 
                                       (+ 1 (book:position book))))
                (g-timeout-add (inexact->exact (* timeout next-timeout))
                               timeout-callback))
              (g-timeout-add timeout timeout-callback)))
  #f)

(define (on-speed-scale-value-changed scale)
  (set! *restart-timer* (+ *restart-timer* 1))
  (let ((t *restart-timer*)
        (speed (get-value scale)))
    (unless (< -1 speed 1)
            (let ((timeout (inexact->exact (/ 60000 (abs speed)))))
              (g-timeout-add timeout
                             (lambda () (word-timer t (if (> speed 0) 
                                                          'forward 
                                                          'backward)
                                                    timeout)))))))

(define (on-play-pause-button-clicked button)
  (let* ((paused? (paused *window*))
         (new-label (if paused?
                       "gtk-media-pause" 
                       "gtk-media-play")))
    (set-label button new-label)
    (set! (paused *window*) (not paused?))))

(define (on-forward-button-clicked button)
  (let ((b (open-book *window*)))
    (set! (book:position b) 
          (+ 10 (book:position b)))
    (with-guard *word-number-spin-guard*
                (set-value (word-number-spin *window*) 
                           (force-between 0 
                                          (book:position b) 
                                          (book:number-of-words b))))))

(define (on-rewind-button-clicked button)
  (let ((b (open-book *window*)))
    (set! (book:position b) 
          (- (book:position b) 10))
    (with-guard *word-number-spin-guard*
                (set-value (word-number-spin *window*) 
                           (force-between 0 
                                          (book:position b) 
                                          (book:number-of-words b))))))

(define (on-start-button-clicked button)
  (let ((b (open-book *window*)))
    (set! (book:position b) -1)
    (with-guard *word-number-spin-guard*
                (set-value (word-number-spin *window*) 0))))

(define (on-end-button-clicked button)
  (let ((b (open-book *window*)))
    (set! (book:position b) (- (book:number-of-words b) 1))
    (with-guard *word-number-spin-guard*
                (set-value (word-number-spin *window*) 
                           (book:number-of-words b)))))

(define (on-current-word-number-spin-value-changed spin)
  (unless *word-number-spin-guard*
          (let ((b (open-book *window*)))
            (set! (book:position b) (get-value-as-int spin)))))

(define (on-text-font-set widget)
  (let* ((font-description (pango-font-description-from-string 
                           (get-font-name widget))))
    ; TODO: remove this is once the relevant bug has been fixed:
    ; at the moment, without this the font setting won't work
    ; http://www.mail-archive.com/guile-gtk-general@gnu.org/msg00040.html
    (write font-description (%make-void-port "w"))
    (let ((font-attribute (pango-attr-font-desc-new font-description))
          (attr-list (pango-attr-list-new)))
      (pango-attr-list-insert attr-list font-attribute)
      (set-attributes (word-display *window*) attr-list))))

;; menu bar setup
; actions for menu entries
(define (close-window-action action)
  (quit-program))

(define (open-file-action action)
  (let ((dialog (make <gtk-file-chooser-dialog> 
                  #:title "Open File"
                  #:action 'open)))
    (add-button dialog
                (gtk-stock-id 'open) 
                (genum->value (make <gtk-response-type> #:value 'accept)))
    (show dialog)
    (connect dialog 'response 
             (lambda (dialog response)
               (case (genum->symbol (make <gtk-response-type> #:value response))
                 ((accept)
                  (open-book-file (get-filename dialog) *window*)
               (destroy dialog)))))))

(define (show-about-action action)
  (show *about-window*))

(define (show-preferences-action action)
  (show *preferences-window*))

; main menu definition to go with the xml
(define +main-menu-entries+
  `(("FileMenu" #f "_File")
    ("OpenFile" ,(gtk-stock-id 'open) "_Open..."
     "<control>O"
     "Open a file"
     ,open-file-action)
    ("CloseWindow" ,(gtk-stock-id 'close) "_Close window"
     "<control>W"
     "Close this window"
     ,close-window-action)
    ("EditMenu" #f "_Edit" #f #f)
    ("Preferences" ,(gtk-stock-id 'preferences) "_Preferences" 
     #f 
     "Show preferences dialog"
    ,show-preferences-action)
    ("HelpMenu" #f "_Help")
    ("About" ,(gtk-stock-id 'about) "_About" #f "Show About dialog"
     ,show-about-action)))

(define (show-main-window)
  (let* ((main-xml (glade-xml-new (find-data-file +glade-main-window-file+) 
                                  #f #f))
         (preferences-xml (glade-xml-new
                           (find-data-file +glade-preferences-window-file+)))
         (window-widget (get-widget main-xml +window-name+))
         (window (make <main-window> #:widget window-widget))
         (this-module (resolve-module '(lukutoukkalib ui mainwindow))))
    (define (add-menubar)
      (let ((box (get-widget main-xml +menu-container-name+))
            (ui-manager (make <gtk-ui-manager>))
            (actions (make <gtk-action-group> #:name "menuactions")))
        (add-ui-from-file ui-manager (find-data-file +ui-manager-file+))
        (add-actions actions +main-menu-entries+)
        (insert-action-group ui-manager actions 0)
        (pack-start box (get-widget ui-manager +menu-path+))))
    (set! (word-display window) (get-widget main-xml +word-display-name+))
    (set! (speed-scale window) (get-widget main-xml +speed-scale-name+))
    (set! (control-container window) 
          (get-widget main-xml +control-container-name+))
    (set! (word-count-label window) 
          (get-widget main-xml +word-count-label-name+))
    (set! (word-number-spin window) (get-widget main-xml +word-number-spin+))
    (set! *window* window)
    (set! *about-window* 
          (get-widget (glade-xml-new 
                       (find-data-file +glade-about-window-file+)) 
                      +about-window-name+))
    (set! *preferences-window*
          (get-widget preferences-xml +preferences-window-name+))
    (connect *preferences-window* 'response 
             (lambda (dialog response) (hide dialog)))
    (connect *preferences-window* 'delete-event
             (lambda (dialog . args) (hide dialog) #t))
    (signal-autoconnect main-xml this-module)
    (signal-autoconnect preferences-xml this-module)
    (connect window-widget 'destroy (lambda (w) (quit-program)))
    (add-menubar)
    (show-all window-widget)
    (open-file-if-argument (program-arguments) window)
    (gtk-main)))
