;;;; Copyright 2005 Juri Pakaste
;;;; 
;;;; This file is part of Lukutoukka.
;;;; 
;;;; Lukutoukka is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.
;;;; 
;;;; Lukutoukka is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with Lukutoukka; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(define-module (lukutoukkalib utils)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (ice-9 syncase)
  #:use-module (ice-9 debug)
  #:export (when unless installation-prefix +glade-main-window-file+ 
                 set-installation-prefix
                 +ui-manager-file+ +glade-about-window-file+ 
                 +glade-preferences-window-file+
                 +glade-error-window-file+
                 ensure-user-data-directory
                 get-bookmark-file
                 find-data-file with-guard assert
                 force-between
                 absolute-path
                 with-output-to-user-file
                 with-input-from-user-file))

(define-macro (when condition . block) 
  (list 'if condition 
        (cons 'begin block)))

(define-macro (unless condition . block)
  (list 'if condition
        #f
        (cons 'begin block)))

 ;; 'assert' comes from Rolf-Thomas Hoppe's 'krims' package at 
 ;; http://www.scsh.net/resources/sunterlib.html
 (define-syntax assert
	(syntax-rules ()
	((assert ?x ?y0 ...)
		(if (not ?x) (error "Assertion failed" '?x ?y0 ...))) ))

; TODO: should deal with exceptions
(define-macro (with-guard guard . block)
  `(begin
     (set! ,guard #t)
     ,(cons 'begin block)
     (set! ,guard #f)))

;; runtime environment helpers

; external constants
(define +glade-main-window-file+ "reader_window.glade")
(define +glade-about-window-file+ "about.glade")
(define +glade-error-window-file+ "error.glade")
(define +glade-preferences-window-file+ "preferences.glade")
(define +ui-manager-file+ "reader-ui.xml")

; internal constants
(define +environment-prefix+ "LUKUTOUKKA_PREFIX")
(define +data-directory+ "/share/lukutoukka/")
(define +user-data-directory+ (string-append (getenv "HOME") "/.lukutoukka"))

(define +installation-prefix+ #f)

(define (get-bookmark-file)
  (let ((filename (string-append (ensure-user-data-directory) "/bookmarks")))
    (if (not (access? filename F_OK))
        (with-output-to-user-file filename
                                  (lambda () (write '()))))
    filename))

; TODO: urgh, think some more about this. at least it should make sure to
; always close those ports - does with-output-to-port deal with exceptions 
; somehow? 
(define (with-output-to-user-file filename thunk)
  (let ((lock-file-port (open-file filename "a")))
    (if (catch #t
                 (lambda () 
                   (flock lock-file-port (logior LOCK_EX LOCK_NB)) 
                   #t)
                 (lambda (key . args) 
                   (format (current-error-port) 
                           "couldn't lock the file for output: ~s ~s~%" 
                           key args) 
                   #f))
        (begin
          (let* ((temp-filename (string-append filename "-XXXXXX"))
                 (temp-file-port (mkstemp! temp-filename))
                 (ans (with-output-to-port temp-file-port thunk)))
            (close-port temp-file-port)
            (rename-file temp-filename filename)
            (close-port lock-file-port)
            ans))
        (begin
          (close-port lock-file-port)
          (throw 'file-locked)))))

(define (with-input-from-user-file filename thunk)
  (let ((file-port (open-file filename "r")))
    (if (catch #t
               (lambda () 
                 (flock file-port (logior LOCK_SH LOCK_NB)) 
                 #t)
               (lambda (key . args) 
                 (format (current-error-port) 
                         "couldn't lock the file for input: ~s ~s~%" key args) 
                 #f))
        (begin
          (let ((ans (with-input-from-port file-port thunk)))
            (close-port file-port)
            ans))
        (begin
          (close-port file-port)
          (throw 'file-locked)))))

(define (set-installation-prefix location)
  (set! +installation-prefix+ location))

(define (find-data-file f)
  (let ((p +installation-prefix+))
    (if (string? p)
        (string-append p +data-directory+ f)
        (error "Couldn't find installation directory"))))
  
(define (ensure-user-data-directory)
  (if (access? +user-data-directory+ F_OK)
      (if (eq? 'directory (stat:type (stat +user-data-directory+)))
          (if (access? +user-data-directory+ (logior R_OK X_OK))
              +user-data-directory+
              (throw 'no-access-to-data-directory
                     "permissions don't allow access to ~/.lukutoukka"))
          (throw 'data-directory-not-a-directory 
                 "~/.lukutoukka exists but is not a directory"))
      (begin
        (mkdir +user-data-directory+)
        +user-data-directory+)))

(define (force-between min-value value max-value)
  "Force value to be between min-value and max-value, inclusive."
  (max min-value (min value max-value)))

; all the characters except /
(define +path-split-character-set+ 
  (char-set-difference char-set:full (char-set #\/)))

(define (absolute-path path)
  "Return the parameter path in absolute form (relative to the root of the file system) and remove any extra slashes, dots or double dots."
  (string-join (reverse! (fold 
                          (lambda (part new-parts)
                            (cond ((or (string= "" part) (string= "." part))
                                   new-parts)
                                  ((or (string= ".." part))
                                   (cdr new-parts))
                                  (else (cons part new-parts))))
                          '()
                          (string-tokenize 
                           (if (string-prefix? "/" path)
                               path
                               (string-append (getcwd) "/" path))
                           +path-split-character-set+)))
               "/"
               'prefix))
